import { Component, ViewChild, DoCheck } from "@angular/core";

import { cLog, Message } from "jalhyd";

import { LogComponent } from "../log/log.component";
import { CalculatorResults } from "../../results/calculator-results";
import { VariableResultsSelectorComponent } from "../variable-results-selector/variable-results-selector.component";
import { I18nService } from "../../services/internationalisation.service";
import { VerificateurResults } from "../../results/verificateur-results";
import { ResultsComponentDirective } from "../fixedvar-results/results.component";
import { LogDrawerComponent } from "../log-drawer/log-drawer.component";

@Component({
    selector: "verificateur-results",
    templateUrl: "./verificateur-results.component.html",
    styleUrls: [
        "./verificateur-results.component.scss"
    ]
})
export class VerificateurResultsComponent extends ResultsComponentDirective implements DoCheck {

    /** résultats non mis en forme */
    private _verificateurResults: VerificateurResults;

    /** true si les résultats doiventt être remis à jour */
    private _doUpdate = false;

    @ViewChild(VariableResultsSelectorComponent)
    private variableResultsSelectorComponent: VariableResultsSelectorComponent;

    @ViewChild("generalLog")
    private generalLogComponent: LogComponent;

    @ViewChild("iterationLog")
    private iterationLogComponent: LogDrawerComponent;

    constructor(
        private i18nService: I18nService,
    ) {
        super();
    }

    public set results(rs: CalculatorResults[]) {
        this._verificateurResults = undefined;
        if (rs.length > 0 && rs[0] instanceof VerificateurResults) {
            this._verificateurResults = rs[0] as VerificateurResults;
        }
        this.updateView();
    }

    /**
     * update iteration log when the variable index changed (event sent by
     * VariableResultsSelectorComponent); variable index is already set in
     * verificateurResults at this time
     */
    public variableIndexChanged() {
        this.updateView();
    }

    public updateView() {
        if (this.iterationLogComponent) {
            this.iterationLogComponent.log = undefined;
        }
        if (this.generalLogComponent) {
            this.generalLogComponent.log = undefined;
        }
        if (this.variableResultsSelectorComponent) {
            this.variableResultsSelectorComponent.results = undefined;
        }
        // set _doUpdate flag so that results are rebuilt on the next Angular display cycle
        this._doUpdate = false;
        if (this._verificateurResults !== undefined) {
            this._doUpdate = this._doUpdate || this._verificateurResults.hasResults || this._verificateurResults.hasLog;
        }
    }

    public ngDoCheck() {
        if (this._doUpdate) {
            this._doUpdate = !this.updateResults();
        }
    }

    /*
     * Retourne les messages à afficher dans le composant de log global, au dessus
     * du sélecteur d'itération : messages globaux du Verificateur
     */
    private get globalLog(): cLog {
        const l = new cLog();
        if (this._verificateurResults && this._verificateurResults.result && this._verificateurResults.result.hasGlobalLog()) {
            l.addLog(this._verificateurResults.result.globalLog);
        }
        return l;
    }

    /**
     * Retourne les messages à afficher dans le composant de log "du bas" : logs de l'itération
     * en cours (messages non-globaux du Vérificateur et eds Espèce), que le résultat varie ou non
     */
    private get iterationLog(): Array<{ message: Message, subLog: cLog }> {
        const l: Array<{ message: Message, subLog: cLog }> = [];
        if (this._verificateurResults) {
            // = 0 lorsque rien ne varie
            const vi = this._verificateurResults.variableIndex;
            if (
                this._verificateurResults.result
                && this._verificateurResults.result.hasResultElements()
                && this._verificateurResults.result.resultElements[vi]
                && this._verificateurResults.result.resultElements[vi].hasLog()
            ) {
                // every log message of Verificateur is supposed to always be a status message for the nth Espece
                const verLog: cLog = this._verificateurResults.result.resultElements[vi].log;
                for (const vm of verLog.messages) {
                    l.push({
                        message: vm,
                        subLog: new cLog()
                    });
                }
            }
            // logs des Espece pour l'itération en cours; l.length devrait toujours être égal à especeResults.length,
            // sauf s'il y a une erreur dans la passe calculée
            if (l.length === this._verificateurResults.especeResults.length) {
                for (let i = 0; i < this._verificateurResults.especeResults.length; i++) {
                    const er = this._verificateurResults.especeResults[i];
                    if (
                        er
                        && er.hasResultElements()
                        && er.resultElements[vi]
                        && er.resultElements[vi].hasLog()) 
                    {
                        l[i].subLog.addLog(er.resultElements[vi].log);
                    }
                }
            } // else silent fail
        }
        return l;
    }

    /**
     * met à jour l'affichage des résultats
     * @returns true si les résultats ont pu être mis à jour
     */
    private updateResults() {
        let selectorUpdated: boolean;

        // results or not, there might be a log
        const logUpdated = (this.iterationLogComponent !== undefined || this.generalLogComponent !== undefined); // gne ?
        if (logUpdated) {
            // order of logs is important !
            this.iterationLogComponent.log = this.iterationLog;
            this.generalLogComponent.log = this.globalLog;
        }

        if (this.hasResults) {
            selectorUpdated = this.variableResultsSelectorComponent !== undefined;
            if (selectorUpdated) {
                this.variableResultsSelectorComponent.results = this._verificateurResults;
            }
        } else {
            selectorUpdated = true;
        }

        return logUpdated && selectorUpdated;
    }

    public get verificateurResults() {
        return this._verificateurResults;
    }

    public get hasResults(): boolean {
        return this._verificateurResults && this._verificateurResults.hasResults;
    }

    public get uitextGeneralLogTitle(): string {
        return this.i18nService.localizeText("INFO_TITREJOURNAL_GLOBAL");
    }

}
